# FuzzyOcr plugin, version 2.1b
# Changelog:
#    version 2.0
#	Replaced imagemagick with netpbm
#	Invoke giffix to fix broken gifs before conversion
#	Support png images
#	Analyze the file to detect the format without content-type
#	Added several configuration parameters
#    version 2.1
#       Added scoring for wrong content-type
#       Added scoring for broken gif images
#       Added configuration for helper applications
#       Added autodisable_score feature to disable the OCR engine if the message has already enough points
#    version 2.1b
#       Rule bugfix to avoid warnings
#    version 2.1c
#       Applied patch provided by Howard Kash to fix problems with spamassassin + Mailscanner + FuzzyOcr
#       Removed '-' from jpegtopnm arguments to provide backwards compatibility for older netpbm versions
#       Fixed typo (treshold -> threshold)
#
# written by Christian Holler decoder_at_own-hero_dot_net

package FuzzyOcr;

use strict;
use Mail::SpamAssassin;
use Mail::SpamAssassin::Util;
use Mail::SpamAssassin::Plugin;

use String::Approx 'adistr';
use IPC::Open2;

our @ISA = qw (Mail::SpamAssassin::Plugin);

our @words = ( );
our $self;
our $pms;

# Default values
our $threshold = "0.3";
our $base_score = "4";
our $add_score = "1";
our $wctypescore = "1.5";
our $cimgscore = "3";
our $countreq = 2;
our $verbose = 1;
our $tmppath = "/tmp";
our $giffix = "/usr/bin/giffix";
our $giftopnm = "/usr/bin/giftopnm";
our $jpegtopnm = "/usr/bin/jpegtopnm";
our $pngtopnm = "/usr/bin/pngtopnm";
our $gocr = "/usr/bin/gocr";
our $dscore = 50;

# constructor: register the eval rule
sub new {
   my ( $class, $mailsa ) = @_;
   $class = ref($class) || $class;
   my $self = $class->SUPER::new($mailsa);
   bless( $self, $class );
   $self->register_eval_rule("check_fuzzy_ocr");
   $self->register_eval_rule("dummy_check");
   return $self;
}

sub parse_config {
  my ($self, $opts) = @_;
  if ($opts->{key} eq "focr_word") {
	push(@words, $opts->{value});
  } elsif ($opts->{key} eq "focr_threshold") {
   	$threshold = $opts->{value};
  } elsif ($opts->{key} eq "focr_base_score") {
        $base_score = $opts->{value};
  } elsif ($opts->{key} eq "focr_add_score") {
        $add_score = $opts->{value};
  } elsif ($opts->{key} eq "focr_corrupt_score") {
        $cimgscore = $opts->{value};
  } elsif ($opts->{key} eq "focr_wrongctype_score") {
        $wctypescore = $opts->{value};
  } elsif ($opts->{key} eq "focr_counts_required") {
        $countreq = $opts->{value};
  } elsif ($opts->{key} eq "focr_verbose") {
  	$verbose = $opts->{value};
  } elsif ($opts->{key} eq "focr_tmp_path") {
  	$tmppath = $opts->{value};
  } elsif ($opts->{key} eq "focr_bin_giffix") {
        $giffix = $opts->{value};
  } elsif ($opts->{key} eq "focr_bin_giftopnm") {
        $giftopnm = $opts->{value};
  } elsif ($opts->{key} eq "focr_bin_jpegtopnm") {
        $jpegtopnm = $opts->{value};
  } elsif ($opts->{key} eq "focr_bin_pngtopnm") {
        $pngtopnm = $opts->{value};
  } elsif ($opts->{key} eq "focr_bin_gocr") {
        $gocr = $opts->{value};
  } elsif ($opts->{key} eq "focr_autodisable_score") {
          $dscore = $opts->{value};
  }
}

sub dummy_check {
	return 0;
}

sub wrong_ctype {
   my ( $format, $ctype ) = @_;
   if ($wctypescore) {
         my $debuginfo = "";
         if ($verbose > 0) {
                $debuginfo = ("\nImage has format \"$format\" but content-type is \"$ctype\"");
         }
         for my $set (0..3) {
            $pms->{conf}->{scoreset}->[$set]->{"FUZZY_OCR_WRONG_CTYPE"} = sprintf("%0.3f", $wctypescore);
         }
         $pms->_handle_hit("FUZZY_OCR_WRONG_CTYPE", $wctypescore, "BODY: ", $pms->{conf}->{descriptions}->{FUZZY_OCR_WRONG_CTYPE} . $debuginfo);
   }
}

sub corrupt_img {
   my ( $err ) = @_;
   if ($cimgscore) {
         my $debuginfo = "";
         if ($verbose > 0) {
                $debuginfo = ("\nCorrupt image: $err");
         }
         for my $set (0..3) {
            $pms->{conf}->{scoreset}->[$set]->{"FUZZY_OCR_CORRUPT_IMG"} = sprintf("%0.3f", $cimgscore);
         }
	 $pms->_handle_hit("FUZZY_OCR_CORRUPT_IMG", $cimgscore, "BODY: ", $pms->{conf}->{descriptions}->{FUZZY_OCR_CORRUPT_IMG} . $debuginfo);
   }
}

sub check_fuzzy_ocr {
   ( $self, $pms ) = @_;
   my @found = ( );
   my $image_type = 0;
   my $cnt = 0;
      
   foreach my $p ( $pms->{msg}->find_parts("image") ) {
      my $cscore = $pms->get_score();
      if ($cscore > $dscore) {
         return 0;
      }
      my ( $ctype, $boundary, $charset, $name ) =
        Mail::SpamAssassin::Util::parse_content_type(
         $p->get_header('content-type') );
	 if ($ctype =~ /image/i) {
	 my $firstline = ($p->decode())[0];
	 my $tempfile = $tmppath . "/" . "spamassassin.$$.focr";
	 my $errfile = $tmppath . "/" . "spamassassin.$$.focr.err";
	 if ($firstline =~ /^\x47\x49\x46/) {
	        if ($ctype !~ /gif/i) {
			wrong_ctype("GIF", $ctype);
		}
	 	$image_type = 1;
		open IMAGE_PROCESSOR, "|$giffix 2>$errfile | $giftopnm - |$gocr -i - > $tempfile";
	 } elsif ($firstline =~ /^\xff\xd8/) {
                if ($ctype !~ /jpeg/i) {
                        wrong_ctype("JPEG", $ctype);
                }
	 	$image_type = 2;
	 	open IMAGE_PROCESSOR, "|$jpegtopnm |$gocr -i - > $tempfile";
	 } elsif ($firstline =~ /^\x89\x50\x4e\x47/) {
                if ($ctype !~ /png/i) {
                        wrong_ctype("PNG", $ctype);
                }
	 	$image_type = 3;
	 	open IMAGE_PROCESSOR, "|$pngtopnm - |$gocr -i - > $tempfile";
	 } else {
	 	$image_type = 0;
		next;
	 }
         foreach $p ( $p->decode() ) {
            print IMAGE_PROCESSOR $p;
         }
	 if ($verbose > 1) {
	 	open DEBUG, ">debug.$$.focr";
	 	print DEBUG "File type: $image_type\n\n"
	 }
	 close IMAGE_PROCESSOR;
	 if (-r $errfile) {
		open ERRFILE, "<$errfile";
		while (<ERRFILE>) {
			if ($_ =~ /GIF-LIB error/i) {
				corrupt_img($_);
				last;
			}
		}
		close ERRFILE;
		unlink $errfile;
	 }
	 open OCR_DATA, "<$tempfile";
	 while (<OCR_DATA>) {
	    s/[^a-zA-Z ]//g;
	    $_ = lc;
	    if ($verbose > 1) {
	    	print DEBUG $_;
	    }
            my $w;
            foreach $w (@words) {
                $w = lc $w;
                my $matched = adistr($w, $_);
                if (abs($matched) < $threshold) {
                        $cnt++;
			if ($verbose > 0) {
                        	push(@found, "\"$w\"" . " with fuzz of " . abs($matched));
			}
                }
            }
         }
	 close OCR_DATA;
	 unlink $tempfile;
	 if ($verbose > 1) {
	 	close DEBUG;
	 }
      }
   }
   if ($cnt >= $countreq) {
   	 my $score = $base_score + ($cnt - $countreq) * $add_score;
	 my $debuginfo = "";
	 if ($verbose > 0) {
	 	$debuginfo = ("\nWords found:\n" . join("\n", @found) ."\n($cnt word occurrences found)");
	 }
         for my $set (0..3) {
            $pms->{conf}->{scoreset}->[$set]->{"FUZZY_OCR"} = sprintf("%0.3f", $score);
         }
         $pms->_handle_hit("FUZZY_OCR", $score, "BODY: ", $pms->{conf}->{descriptions}->{FUZZY_OCR} . $debuginfo);
   }
   return 0;
}

1;
